# pylint: skip-file
import pytest
from pathlib import Path
from suspect.pyomo.util import (
    model_variables,
    model_objectives,
    model_constraints,
)
from suspect.pyomo import read_osil


current_dir = Path(__file__).resolve().parent


def test_osil_read():
    filename = current_dir / 'osil' / 'example1.xml'
    problem = read_osil(str(filename))
    variables = list(model_variables(problem))
    constraints = list(model_constraints(problem))
    objectives = list(model_objectives(problem))
    assert len(variables) == 35
    assert len(constraints) == 18
    assert len(objectives) == 1



# Copyright 2018 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Convexity detection rules for quadratic expressions."""
import numpy as np
from suspect.convexity.convexity import Convexity
from suspect.convexity.rules.rule import ConvexityRule



class QuadraticRule(ConvexityRule):
    """Return convexity of quadratic."""

    def apply(self, expr, convexity, mono, bounds):
        # Sum of squares
        if self._is_sum_of_squares(expr):
            coefficients = np.array([term.coefficient for term in expr.terms])
            if np.all(coefficients >= 0):
                return Convexity.Convex
            elif np.all(coefficients <= 0):
                return Convexity.Concave
            else:
                return Convexity.Unknown

        # try compute eigvalues
        n = len(expr.args)
        A = np.zeros((n, n))
        var_to_idx = dict([(v, i) for i, v in enumerate(expr.args)])
        for term in expr.terms:
            i = var_to_idx[term.var1]
            j = var_to_idx[term.var2]
            if term.var1 == term.var2:
                A[i, j] = term.coefficient
            else:
                A[i, j] = term.coefficient / 2.0
                A[j, i] = A[i, j]
        eigv = np.linalg.eigvalsh(A)
        if np.all(eigv >= 0):
            return Convexity.Convex
        elif np.all(eigv <= 0):
            return Convexity.Concave
        return Convexity.Unknown

    def _is_sum_of_squares(self, expr):
        if len(expr.terms) == 0:
            return False
        for term in expr.terms:
            if term.var1 != term.var2:
                return False
        return True



# pylint: skip-file
import pytest
import pyomo.environ as aml
from tests.conftest import PlaceholderExpression
from suspect.pyomo import dag_from_pyomo_model
from suspect.dag.dag import ProblemDag
from suspect.dag.iterator import DagForwardIterator, DagBackwardIterator
import suspect.pyomo.expressions as pex
from hypothesis import given, assume
import hypothesis.strategies as st


def create_problem():

    # build a dag that looks like this:
    #
    #            c0    c1
    #            |     |
    #            |     |
    #            +    |.|
    #            | \   |
    #            |   \ |
    #            S     ^
    #          /  \   / \
    #        x0     x1   c

    m = aml.ConcreteModel()

    m.x0 = aml.Var()
    m.x1 = aml.Var()

    m.c0 = aml.Constraint(expr=m.x0 - m.x1 + m.x1 ** 3 <= 0)
    m.c1 = aml.Constraint(expr=abs(m.x1 ** 3) <= 10.0)

    return dag_from_pyomo_model(m)


@pytest.fixture
def problem_dag():
    return create_problem()


class FakeForwardVisitor(object):
    def __init__(self, visit_returns):
        self.seen = set()
        self._visit_returns = visit_returns

    def visit(self, expr, ctx):
        self.seen.add(id(expr))
        return self._visit_returns(expr)


class TestDagForwardIterator(object):
    def test_no_starting_point(self, problem_dag):
        it = DagForwardIterator()
        visitor = FakeForwardVisitor(lambda v: True)
        changes = it.iterate(problem_dag, visitor, None)
        assert changes[0].is_variable_type()
        assert changes[-1].is_sink
        assert len(changes) == 10

    def test_early_stop(self, problem_dag):
        it = DagForwardIterator()
        visitor = FakeForwardVisitor(
            lambda v: False if isinstance(v, pex.PowExpression) else True,
        )
        changes = it.iterate(problem_dag, visitor, None)
        assert len(changes) == 10-3

    def test_starting_point(self, problem_dag):
        it = DagForwardIterator()
        visitor = FakeForwardVisitor(lambda v: False)
        changes = it.iterate(problem_dag, visitor, None, starting_vertices=[])
        assert len(visitor.seen) == 4
        assert len(changes) == 0


class FakeBackwardVisitor(object):
    def __init__(self, visit_returns):
        self.seen = {}
        self._visit_returns = visit_returns

    def visit(self, expr, ctx):
        self.seen[id(expr)] = expr
        return self._visit_returns(expr)


class TestDagBackwardIterator(object):
    def test_no_starting_point(self, problem_dag):
        it = DagBackwardIterator()
        visitor = FakeBackwardVisitor(lambda v: True)
        changes = it.iterate(problem_dag, visitor, None)
        assert len(changes) == 10
        assert changes[0].is_sink
        assert changes[-1].is_variable_type()

    def test_early_stop(self, problem_dag):
        it = DagBackwardIterator()
        visitor = FakeBackwardVisitor(
            lambda v: False if isinstance(v, pex.SumExpression) else True,
        )
        changes = it.iterate(problem_dag, visitor, None)
        assert len(changes) == 10-4

    def test_starting_point(self, problem_dag):
        it = DagBackwardIterator()
        visitor = FakeBackwardVisitor(lambda v: False)
        changes = it.iterate(problem_dag, visitor, None, starting_vertices=[])
        assert len(visitor.seen) == 2
        assert len(changes) == 0
